import logging
import os
import socket
import sys
import tarfile
import threading
from optparse import OptionParser
from subprocess import PIPE, Popen

errorList = []

# Update these locations as per requirements
# Default log location will fetch everything and might increase support bundle size
LOG_LOCATIONS = "/var/log/pods, /var/log/containers, /var/log/calico, /var/lib/docker, /var/lib/nfs, /var/lib/kubelet"
CONFIG_LOCATIONS = "/etc/docker, /etc/kubernetes"

# System namespace objects
NAMESPACES_SYSTEM = "kube-system, vmware-system-csi"
RESOURCES_SYSTEM = "pod, statefulset, deployment, daemonset, configmap, secret"

# User namespaces objects. Currently just default is being considered as user namespaces
NAMESPACES_USER = ""
RESOURCES_USER = "pvc, pod, statefulset, deployment"

# Cluster objects
RESOURCES_CLUSTER = "node, pv, volumeattachment"


def GetOptions():
    """
    Supports the command-line arguments listed below
    """
    parser = OptionParser()
    parser.add_option("-d", "--dest", default="/tmp/",
                      help="Path to store log files")
    parser.add_option("-t", "--nodeType", default="",
                      help="Node type in use")
    (options, _) = parser.parse_args()
    return options


def execShellCommand(command):
    """
    Executes command on localhost
    """
    try:
        process = Popen(args=command, stdout=PIPE, stderr=PIPE, shell=True)
        stdout, stderr = process.communicate()
        stdout = str(stdout, 'utf-8')
        stderr = str(stderr, 'utf-8')
        return stdout, stderr
    except Exception as e:
        print("Failed to execute command \'%s\'. Error: %s", command, str(e))


def make_tarfile(output_filename, source_dirs):
    """
    Creates tar file for provided directory tree
    """
    TAR_COMMAND = "tar --ignore-failed-read --exclude='/var/lib/docker/overlay2' -iczhf {} ".format(
        output_filename)
    for directory in source_dirs:
        directory = directory.strip()
        sys.stdout.write(
            "Collecting logs from {} ... \n".format(directory))
        # with tarfile.open(output_filename, "w:gz") as tar:
        #     tar.add(directory, arcname=os.path.basename(basename))
        if output_filename == directory[:-1] + ".tgz":
            TAR_COMMAND = TAR_COMMAND + " -C {} . ".format(directory)
        else:
            TAR_COMMAND = TAR_COMMAND + directory + " "

    print(TAR_COMMAND)
    _, stderr = execShellCommand(str(TAR_COMMAND))
    if stderr != "":
        errorList.append(
            "Error creating tar file for support bundle: {}".format(stderr.strip()))


def createSupportDirectory():
    global k8s_support_dir_path
    k8s_support_dir_name = str(socket.gethostname())
    _, stderr = execShellCommand(
        str("mkdir -p " + options.dest + k8s_support_dir_name))
    k8s_support_dir_path = str(options.dest + k8s_support_dir_name + "/")
    if stderr != "":
        errorList.append("Error creating support bundle: {}".format(stderr))
    return k8s_support_dir_name


def collectKubeletLogs():
    """
    Collects kubelet Logs from all nodes(master + workers) in k8s cluster and stores in sub-directory kubelet   
    """
    global kubeletLogsDestDirectoryPath
    kubeletLogsDestDirectoryPath = str(k8s_support_dir_path + "kubelet/")

    try:
        # Create directory to collect kubelet logs
        execShellCommand(str("mkdir -p " + kubeletLogsDestDirectoryPath))

        kubeletLogCommand = str("journalctl -xeu kubelet > " +
                                kubeletLogsDestDirectoryPath + options.nodeType + "-" + str(socket.gethostname()) + "-kubelet.log")
        execShellCommand(kubeletLogCommand)
    except Exception as e:
        print("Failed to collect kubelet logs due to : %s", str(e))
        log.error('Failed to collect kubelet logs. Error : %s', str(e))


def runNamespaceResourceDescribe(resourceObj, resourceNamespace, outputPath):
    """
    Runs kubectl describe on provided rescource and namespace in required output path
    """
    try:
        command = "kubectl get {} -n {} | awk '{{print $1}}' | tail -n +2".format(
            resourceObj, resourceNamespace)
        stdout_resources, stderr_resources = execShellCommand(
            command)
        if stderr_resources != "":
            if "No" in str(stderr_resources):
                errorList.append(
                    "Error getting " + resourceObj + " names: {}".format(stderr_resources.strip()))
                return None
        resources = stdout_resources.split("\n")
    except Exception as e:
        print("Failed to get resources. Error: {}".format(str(e)))
        return None

    for res in resources:
        try:
            RESOURCE_DESCRIBE_COMMAND = "kubectl describe {} {} -n {} >> {}_describe.txt".format(
                resourceObj, res, resourceNamespace, outputPath + resourceObj)
            _, stderr_resource_describe = execShellCommand(
                RESOURCE_DESCRIBE_COMMAND)
            if stderr_resource_describe != "":
                errorList.append(
                    "Error describing " + res + " : {}".format(stderr_resource_describe))
        except Exception as e:
            print("Failed to get describe of %s %s. Error: %s",
                  res, resourceObj, str(e))


def runClusterResourceDescribe(resourceObj, outputPath):
    """
    Runs kubectl describe on provided cluster rescource in required output path
    """
    try:
        command = "kubectl get {} | awk '{{print $1}}' | tail -n +2".format(
            resourceObj)
        stdout_resources, stderr_resources = execShellCommand(
            command)
        if stderr_resources != "":
            if "No" in str(stderr_resources):
                errorList.append(
                    "Error getting " + resourceObj + " names: {}".format(stderr_resources.strip()))
                return None
        resources = stdout_resources.split("\n")
    except Exception as e:
        print("Failed to get resources. Error: {}".format(str(e)))
        return None

    for res in resources:
        try:
            RESOURCE_DESCRIBE_COMMAND = "kubectl describe {} {} >> {}_describe.txt".format(
                resourceObj, res, outputPath + resourceObj)
            _, stderr_resource_describe = execShellCommand(
                RESOURCE_DESCRIBE_COMMAND)
            if stderr_resource_describe != "":
                errorList.append(
                    "Error describing " + res + " : {}".format(stderr_resource_describe))
        except Exception as e:
            print("Failed to get describe of %s %s. Error: %s",
                  res, resourceObj, str(e))


def getObjectEvents(objectType, objectOutputPath, resourceTypes):
    """
    Collects cluster resource describe events and stores at sub-directory describe_events
    """
    try:
        commandsDestDirectoryPath = str(
            objectOutputPath + "describe_events/")

        # Create directory to collect cluster resource events
        execShellCommand(str("mkdir -p " + commandsDestDirectoryPath))
    except Exception as e:
        print("Failed to create support sub-directory describe_events. Error: %s", str(e))
        return None

    for res in resourceTypes.split(","):
        res = res.strip()
        sys.stdout.write(
            "Collecting describe events of all {} in {} ...  \n".format(res, objectType))

        # Collect resource describe events
        if objectType is "cluster":
            runClusterResourceDescribe(res, commandsDestDirectoryPath)
        else:
            runNamespaceResourceDescribe(
                res, objectType, commandsDestDirectoryPath)


def runClusterCommands():
    """
    Runs various commands to collect cluster Information and stores in sub-directory commands
    """
    try:
        commandsDestDirectoryPath = str(k8s_support_dir_path + "commands/")

        # Create directory to collect cluster command outputs
        execShellCommand(str("mkdir -p " + commandsDestDirectoryPath))

        # Create directory to collect cluster objects outputs
        clusterOutpuDestDirectoryPath = str(
            commandsDestDirectoryPath + "cluster/")
        execShellCommand(str("mkdir -p " + clusterOutpuDestDirectoryPath))

        # Set kubeconfig to run kubectl commands
        execShellCommand(str(
            "mkdir -p $HOME/.kube && sudo cp /etc/kubernetes/admin.conf $HOME/.kube/config && sudo chown $(id -u):$(id -g) $HOME/.kube/config"))

        # Collect resource info of cluster objects
        for res in RESOURCES_CLUSTER.split(","):
            res = res.strip()
            sys.stdout.write(
                "Collecting {} cluster resource info ... \n".format(res))

            # Collects detailed info on cluster resource
            kubeSystemPodInfoCommand = str("kubectl get {} -o yaml > {}{}.yaml".format(res,
                                                                                       clusterOutpuDestDirectoryPath, res))
            execShellCommand(kubeSystemPodInfoCommand)

            # Collects minimum required info on all system pods in a cluster
            kubeSystemPodInfoCommand = str("kubectl get {} -o wide > {}{}.min.txt".format(res,
                                                                                          clusterOutpuDestDirectoryPath, res))
            execShellCommand(kubeSystemPodInfoCommand)

        # Collect describe events for namespace objects
        getObjectEvents(
            "cluster", clusterOutpuDestDirectoryPath, RESOURCES_CLUSTER)

        for ns in NAMESPACES_SYSTEM.split(","):
            ns = ns.strip()
            # Create directory to collect system namespace outputs
            namespaceDestDirectoryPath = str(
                commandsDestDirectoryPath + ns + "/")
            execShellCommand(str("mkdir -p " + namespaceDestDirectoryPath))

            # Collect resource info in current namespace context
            for res in RESOURCES_SYSTEM.split(","):
                res = res.strip()
                sys.stdout.write(
                    "Collecting {} resource info from {} namespace ... \n".format(res, ns))

                # Collects detailed info on all system pods in a cluster
                kubeSystemPodInfoCommand = str("kubectl get {} -n {} -o yaml > {}{}.yaml".format(res,
                                                                                                 ns, namespaceDestDirectoryPath, res))
                execShellCommand(kubeSystemPodInfoCommand)

                # Collects minimum required info on all system pods in a cluster
                kubeSystemPodInfoCommand = str("kubectl get {} -n {} -o wide > {}{}.min.txt".format(res,
                                                                                                    ns, namespaceDestDirectoryPath, res))
                execShellCommand(kubeSystemPodInfoCommand)

            # Collect describe events for namespace objects
            getObjectEvents(ns, namespaceDestDirectoryPath, RESOURCES_SYSTEM)

        default_user_namespace = "default"
        # Create directory to collect system namespace outputs
        namespaceDestDirectoryPath = str(
            commandsDestDirectoryPath + default_user_namespace + "/")
        execShellCommand(str("mkdir -p " + namespaceDestDirectoryPath))

        for res in RESOURCES_USER.split(","):
            res = res.strip()
            sys.stdout.write(
                "Collecting {} resource info from {} namespace ... \n".format(res, default_user_namespace))

            # Collects detailed info of resources in cluster
            resInfoCommand = str("kubectl get {} -o yaml > {}{}.yaml".format(
                res, namespaceDestDirectoryPath, res))
            execShellCommand(resInfoCommand)

            # Collects minimum required info of resources in a cluster
            resInfoCommand = str("kubectl get {} -o wide > {}{}.min.txt".format(
                res, namespaceDestDirectoryPath, res))
            execShellCommand(resInfoCommand)

        # Collect describe events for namespace objects
        getObjectEvents(default_user_namespace,
                        namespaceDestDirectoryPath, RESOURCES_USER)
    except Exception as e:
        print("Failed to run cluster commands on k8s clusters. Error: %s", str(e))


def provideFilePermissions(directory):
    CHMOD_COMMAND = str("sudo chmod -R 755 " + directory)
    execShellCommand(CHMOD_COMMAND)


def main():
    # Process Command line args
    global options
    options = GetOptions()

    # Create Support directory
    global support_dir

    try:
        support_dir = createSupportDirectory()
    except Exception as e:
        print("Failed to Create Support Directory due to : %s", str(e))
        log.error('Failed to create support dirctory. Error : %s', str(e))
        return

    # Collect Cluster Info
    if "worker" not in options.nodeType:
        runClusterCommands()

    # Collect kubelet logs
    collectKubeletLogs()

    k8s_support_dir_paths = []
    k8s_support_dir_paths.append(k8s_support_dir_path)
    k8s_support_dir_paths.extend(LOG_LOCATIONS.split(","))
    k8s_support_dir_paths.extend(CONFIG_LOCATIONS.split(","))

    # Create compressed archive of support bundle
    try:
        generate_support_bundle_process = threading.Thread(
            name='support-bundle-create', target=make_tarfile(options.dest + support_dir + ".tgz", k8s_support_dir_paths))
        generate_support_bundle_process.start()
        screen_info = "Tanzu K8s Cluster Support bundle '{}' generated at '{}'"
        screen_info_log = str(screen_info.format(
            support_dir + ".tgz", options.dest))
        while generate_support_bundle_process.is_alive():
            continue
        sys.stdout.write(screen_info_log)
        log.info(screen_info_log)

        # Give Permissions to support bundle files
        provideFilePermissions(options.dest)
    except Exception as e:
        return


# Start script
if __name__ == "__main__":
    log = logging.getLogger(__name__)
    support_process = threading.Thread(name='tanzu-k8s-support', target=main)
    support_process.start()
    while support_process.is_alive():
        continue
    print("\nComplete..")
    if len(errorList) > 0:
        print("Errors during script exec runtime: \n", errorList)
